/*
 * Copyright (c) 2001 Tamotsu Hattori.
 * All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/sysctl.h>
#include <longrun.h>

static char    *mode_name[] = {
	"min-frequency mode",
	"economy mode",
	"performance mode",
	"max-frequency mode",
	"unknown mode",
	NULL
};

struct longrun_status {
	u_int	frequency;
	u_int	voltage;
	u_int	percentage;
};

struct keytab {
	char           *kt_cp;
	int             kt_i;
}               keywords[] = {
#define K_GETMODE 1
	{"getmode", K_GETMODE},
#define K_GETSTAT 2
	{"getstat", K_GETSTAT},
#define K_MONITOR 3
	{"monitor", K_MONITOR},
#define K_SETMODE 4
	{"setmode", K_SETMODE},
	{0, 0}
};

int 
keyword(char *arg)
{
	struct keytab  *kt = keywords;

	while (kt->kt_cp && strcmp(kt->kt_cp, arg))
		kt++;
	return kt->kt_i;
}

void 
usage(char *arg)
{
	if (arg)
		fprintf(stderr, "bad keyword: %s", arg);

	fprintf(stderr,
		"usage: longrunctl [-c count] [-i interval] command [mode]\n");
	fprintf(stderr,
		"\tcommand: one of the followings: getmode, setmode, getstat, monitor.\n");
	fprintf(stderr,
		"\t\tgetmode: get current longrun mode.\n");
	fprintf(stderr,
		"\t\tsetmode: change longrun mode to specified one.\n");
	fprintf(stderr,
		"\t\tgetstat: get current longrun status such as frequency.\n");
	fprintf(stderr,
		"\t\tmonitor: monitor longrun status.\n");
	fprintf(stderr,
		"\tcount: iteration count.\n");
	fprintf(stderr,
		"\tinterval: interval of reporting status.\n");
	fprintf(stderr,
		"\tmode: longrun mode (0 to 3, 0 is max saving power).\n");
	exit(1);
}

int 
get_mode(void)
{
	u_int	mode;
	size_t	len;

	len = sizeof(mode);
	if (sysctlbyname("hw.crusoe.longrun", &mode, &len, NULL, 0) < 0)
		return -1;

	switch (mode) {
	case LONGRUN_MODE_MINFREQUENCY:
	case LONGRUN_MODE_ECONOMY:
	case LONGRUN_MODE_PERFORMANCE:
	case LONGRUN_MODE_MAXFREQUENCY:
		return (int)mode;
	default:
		return LONGRUN_MODE_UNKNOWN;
	}
}

int 
get_current_stat(struct longrun_status *l_stat)
{
	size_t	len;

	len = sizeof(l_stat->frequency);
	if (sysctlbyname("hw.crusoe.frequency", &(l_stat->frequency), &len, NULL, 0) < 0) 
		return -1;

	len = sizeof(l_stat->voltage);
	if (sysctlbyname("hw.crusoe.voltage", &(l_stat->voltage), &len, NULL, 0) < 0) 
		return -1;

	len = sizeof(l_stat->percentage);
	if (sysctlbyname("hw.crusoe.percentage", &(l_stat->percentage), &len, NULL, 0) < 0) 
		return -1;

	return 0;
}

int
get_stat(void)
{
	struct longrun_status l_stat;

	if (get_current_stat(&l_stat) < 0)
		return -1;
	fprintf(stdout,
		"frequency: %d[MHz]\nvoltage: %.2f[V]\npercentage: %d[%%]\n",
		l_stat.frequency,
		l_stat.voltage / 1000.0,
		l_stat.percentage);
	return 0;
}

int 
monitor(int count, int interval)
{
	struct longrun_status l_stat;
	int             tic;
	if (count < 0 || interval < 0)
		return -1;
	tic = (count != 0);
	count += (count == 0);
	for (; count > 0; count -= tic) {
		if(get_current_stat(&l_stat) < 0)
			return -1;
		fprintf(stdout,
			"%d[MHz], %.2f[V], %3d[%%]\r",
			l_stat.frequency,
			l_stat.voltage / 1000.0,
			l_stat.percentage);
		fflush(stdout);
		sleep(interval);
	}
	putchar('\n');
	return 0;
}

int 
set_mode(u_int newmode)
{
	u_int	oldmode;
	size_t	oldlen, newlen;

	oldlen = sizeof(oldmode);
	newlen = sizeof(newmode);
	if (sysctlbyname("hw.crusoe.longrun", &oldmode, &oldlen, &newmode, newlen) < 0)
		return -1;
	return (int)oldmode;
}

int
main(int argc, char *argv[])
{
	int	ch;
	int	sflag = 0;
	struct longrun_status	l_stat;
	int	mode = 0, count = 0, interval = 1, oldmode;

	if (argc < 2)
		usage((char *)NULL);
	while ((ch = getopt(argc, argv, "i:c:")) != -1)
		switch (ch) {
		case 'i':
			interval = atoi(optarg);
			break;
		case 'c':
			count = atoi(optarg);
			/* optind++; */
			break;
		default:
			usage((char *)NULL);
		}
	argc -= optind;
	argv += optind;

	if (*argv)
		switch (keyword(*argv)) {
		case K_GETMODE:
			mode = get_mode();
			printf("Current mode: %s\n", mode_name[mode]);
			exit(0);
		case K_GETSTAT:
			get_stat();
			exit(0);
		case K_MONITOR:
			monitor(count, interval);
			exit(0);
		case K_SETMODE:
			if (*(argv + 1) == NULL) {
				fprintf(stderr, "no longrun mode to set.\n");
				exit(1);
			}
			mode = atoi(*(argv + 1));
			switch (mode) {
			case LONGRUN_MODE_MINFREQUENCY:
			case LONGRUN_MODE_ECONOMY:
			case LONGRUN_MODE_PERFORMANCE:
			case LONGRUN_MODE_MAXFREQUENCY:
				break;
			default:
				fprintf(stderr, "invalid longrun mode.\n");
				exit(1);
			}
			if ((oldmode = set_mode(mode)) < 0)
				exit(1);
			fprintf(stderr, "Mode changed: %s -> %s\n", mode_name[oldmode], mode_name[mode]);
			exit(0);
		}
	usage((char *)NULL);
}


